/*
    Copyright  2003, The KPD-Team
    All rights reserved.
    http://www.mentalis.org/

  Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions
  are met:

    - Redistributions of source code must retain the above copyright
       notice, this list of conditions and the following disclaimer. 

    - Neither the name of the KPD-Team, nor the names of its contributors
       may be used to endorse or promote products derived from this
       software without specific prior written permission. 

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
  FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL
  THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
  INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
  (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
  STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
  OF THE POSSIBILITY OF SUCH DAMAGE.
*/

using System;
using System.IO;
using System.Text;
using System.Runtime.InteropServices;

namespace Org.Mentalis.Multimedia {
	/// <summary>
	/// Represents a WAVE sound file.
	/// </summary>
	public sealed class WaveFile {
		/// <summary>
		/// Initializes a new instance of the <see cref="WaveFile"/> class.
		/// </summary>
		/// <param name="file">The file to open.</param>
		/// <exception cref="MediaException">An error occurs while reading the media file.</exception>
		public WaveFile(string file) : this(file, 5) {}
		/// <summary>
		/// Initializes a new instance of the <see cref="WaveFile"/> class.
		/// </summary>
		/// <param name="file">The file to open.</param>
		/// <param name="bufferCount">The number of memory buffers to use when reading the file.</param>
		/// <remarks>Setting the number of memory buffer to a low number may cause skips during playback. Setting the number of buffer to a high number may cause excessive memory usage.</remarks>
		/// <exception cref="MediaException">An error occurs while reading the media file.</exception>
		public WaveFile(string file, int bufferCount) {
			if (file == null)
				throw new ArgumentNullException();
			if (bufferCount <= 0)
				throw new ArgumentException();
			m_Filename = file;
			m_Disposed = false;
			m_Buffers = bufferCount;
			Open(file);
			m_Status = MediaStatus.Stopped;
		}
		/// <summary>
		/// Converts byte offsets returned by <see cref="Position"/> or <see cref="Length"/> to milliseconds.
		/// </summary>
		/// <param name="byteOffset">The byte offset the convert.</param>
		/// <returns>The corresponding milliseconds.</returns>
		public int ToMilliSeconds(int byteOffset) {
			return ((byteOffset * m_Format.nBlockAlign) / m_Format.nAvgBytesPerSec) * 1000;
		}
		/// <summary>
		/// Converts byte offsets returned by <see cref="Position"/> or <see cref="Length"/> to a byte offset.
		/// </summary>
		/// <param name="milliseconds">The number of milliseconds to convert.</param>
		/// <returns>The corresponding byte offset.</returns>
		public int ToByteOffset(int milliseconds) {
			return ((milliseconds / 1000) * m_Format.nAvgBytesPerSec) / m_Format.nBlockAlign;
		}
		/// <summary>
		/// Opens a WAVE file and initializes the internal structures.
		/// </summary>
		/// <param name="file">The file to open.</param>
		/// <exception cref="MediaException">An error occurs while reading the media file.</exception>
		private void Open(string file) {
			// Make sure the file exists
			if (!File.Exists(file))
				throw new FileNotFoundException("The media file could not be found.", file);
			// Open the input file
			Input = mmioOpen(file, IntPtr.Zero, MMIO_READ);
			if (Input.Equals(IntPtr.Zero))
				throw new MediaException("An error occurs while opening the media file.");
			// make sure this is a wave file
			MMCKINFO mmi = new MMCKINFO();
			mmi.fccType = mmioStringToFOURCC("WAVE", 0);
			if (mmioDescend(Input, ref mmi, IntPtr.Zero, MMIO_FINDRIFF) != MMSYSERR_NOERROR) {
				Close();
				throw new MediaException("The media file is invalid.");
			}
			// get format info
			mmi = new MMCKINFO();
			mmi.ckid = mmioStringToFOURCC("fmt", 0);
			if (mmioDescend(Input, ref mmi, IntPtr.Zero, MMIO_FINDCHUNK) != MMSYSERR_NOERROR) {
				Close();
				throw new MediaException("The format chunk could not be found.");
			}
			m_Format = new WAVEFORMAT();
			if (mmioRead(Input, ref m_Format, Marshal.SizeOf(m_Format)) == -1) {
				Close();
				throw new MediaException("An error occurs while reading from the media file.");
			}
			mmioAscend(Input, ref mmi, 0);
			// find the data subchunk
			mmi.ckid = mmioStringToFOURCC("data", 0);
			if (mmioDescend(Input, ref mmi, IntPtr.Zero, MMIO_FINDCHUNK) != MMSYSERR_NOERROR) {
				Close();
				throw new MediaException("The data chunk could not be found.");
			}
			m_DataOffset = mmioSeek(Input, 0, SEEK_CUR);
			// Get the length of the audio
			m_AudioLength = mmi.ckSize;
			// Allocate audio buffers
			m_BufferSize = (int)(m_Format.nSamplesPerSec * m_Format.nBlockAlign * m_Format.nChannels * BUFFER_SECONDS);
			m_BufferSize = m_BufferSize - (m_BufferSize % m_Format.nBlockAlign);
			m_Buffer = new IntPtr[m_Buffers];
			m_Header = new WAVEHDR[m_Buffers];
			for (int i = 0; i < m_Buffer.Length; i++) {
				m_Buffer[i] = Marshal.AllocHGlobal(m_BufferSize);
			}
		}
		/// <summary>
		/// Starts playing the media file synchronously.
		/// </summary>
		/// <remarks>This method will block until the file has finished playing.</remarks>
		/// <exception cref="ObjectDisposedException">The file has been closed.</exception>
		/// <exception cref="MediaException">An error occurs while playing the media file.</exception>
		public void Play() {
			IAsyncResult ar = BeginPlay(null, null);
			ar.AsyncWaitHandle.WaitOne();
		}
		/// <summary>
		/// Starts playing the media file asynchronously.
		/// </summary>
		/// <param name="callback">An <see cref="AsyncCallback"/> instance. This delegate will be called if the playback is finished. This member can be a null reference (<b>Nothing</b> in Visual Basic).</param>
		/// <param name="state">An object containing state information for this request.</param>
		/// <returns>An IAsyncResult that references the asynchronous playback.</returns>
		/// <exception cref="ObjectDisposedException">The file has been closed.</exception>
		/// <exception cref="MediaException">An error occurs while playing the media file.</exception>
		public IAsyncResult BeginPlay(AsyncCallback callback, object state) {
			if (m_Disposed)
				throw new ObjectDisposedException(this.GetType().FullName);
			if (m_Status == MediaStatus.Playing)
				return m_AsyncResult;
			if (waveOutOpen(ref m_Output, 0, ref m_Format, new WaveDelegate(WaveCallBack), 0, CALLBACK_FUNCTION) != MMSYSERR_NOERROR)
				throw new MediaException("An error occurs while opening the WAVE device.");
			for (int i = 0; i < m_Header.Length; i++) {
				m_Header[i].lpData = m_Buffer[i];
				m_Header[i].dwBufferLength = m_BufferSize;
				m_Header[i].dwFlags = 0;
				m_Header[i].dwLoops = 0;
				if (waveOutPrepareHeader(Output, ref m_Header[i], Marshal.SizeOf(m_Header[i])) != MMSYSERR_NOERROR)
					throw new MediaException("An error occurs while preparing the WAVE buffers.");
			}
			m_Stopping = -1;
			m_Status = MediaStatus.Playing;
			m_StartPos = mmioSeek(Input, 0, SEEK_CUR) - m_DataOffset;
			for (int i = 0; i < m_Header.Length; i++) {
				WaveCallBack(Input, MM_WOM_DONE, 0, ref m_Header[i], 0);
			}
			m_AsyncResult = new MediaResult(callback, state);
			return m_AsyncResult;
		}
		/// <summary>
		/// Called whenever a data buffer has been played.
		/// </summary>
		/// <param name="hwo">Handle to the waveform-audio device associated with the callback.</param>
		/// <param name="uMsg">Waveform-audio output message.</param>
		/// <param name="dwInstance">User-instance data specified with waveOutOpen.</param>
		/// <param name="wavhdr">Message parameter.</param>
		/// <param name="dwParam2">Message parameter.</param>
		private void WaveCallBack(IntPtr hwo, int uMsg, int dwInstance, ref WAVEHDR wavhdr, int dwParam2) {
			if (uMsg == MM_WOM_DONE && m_Stopping == -1) {
				int rc;
				int remaining  = (m_DataOffset + m_AudioLength - mmioSeek(Input, 0, SEEK_CUR));
				if (m_BufferSize < remaining) {
					rc = mmioRead(Input, wavhdr.lpData, m_BufferSize);
				} else {
					rc = mmioRead(Input, wavhdr.lpData, remaining);
					m_Stopping = 0;
				}
				if (rc != -1) {
					wavhdr.dwBufferLength = rc;
					rc = waveOutWrite(Output, ref wavhdr, Marshal.SizeOf(wavhdr));
				}
			}
			if (uMsg == MM_WOM_DONE && m_Stopping != -1) {
				m_Stopping++;
				if (m_Stopping == m_Buffers + 1) {
					this.Position = 0;
					m_Status = MediaStatus.Stopped;
					m_AsyncResult.Notify();
				}
			}
		}
		/// <summary>
		/// Stops the playback of the file.
		/// </summary>
		/// <exception cref="ObjectDisposedException">The file has been closed.</exception>
		public void Stop() {
			if (m_Disposed)
				throw new ObjectDisposedException(this.GetType().FullName);
			if (m_Status == MediaStatus.Stopped)
				return;
			m_Stopping = 1;
			waveOutReset(Output);
			for (int i = 0; i < m_Header.Length; i++) {
				waveOutUnprepareHeader(Output, ref m_Header[i], Marshal.SizeOf(m_Header[i]));
			}
			waveOutClose(Output);
			Position = 0;
			m_Status = MediaStatus.Stopped;
		}
		/// <summary>
		/// Pauses or unpauses the playback of the file.
		/// </summary>
		/// <value><b>true</b> if the file is paused, <b>false</b> otherwise.</value>
		/// <exception cref="ObjectDisposedException">The file has been closed.</exception>
		public bool Paused {
			get {
				if (m_Disposed)
					throw new ObjectDisposedException(this.GetType().FullName);
				return this.Status == MediaStatus.Paused;
			}
			set {
				if (m_Disposed)
					throw new ObjectDisposedException(this.GetType().FullName);
				if (this.Status == MediaStatus.Stopped)
					return;
				if (value && this.Status == MediaStatus.Playing) {
					waveOutPause(Output);
					m_Status = MediaStatus.Paused;
				} else if (!value && this.Status == MediaStatus.Paused) {
					waveOutRestart(Output);
					m_Status = MediaStatus.Playing;
				}
			}
		}
		/// <summary>
		/// Gets the full path of the media file.
		/// </summary>
		/// <value>A <see cref="String"/>.</value>
		public string Filename {
			get {
				return m_Filename;
			}
		}
		/// <summary>
		/// Gets the length of the media file.
		/// </summary>
		/// <value>An Integer representing the length of the file.</value>
		/// <remarks>You can convert between the byte offset of this property and milliseconds by using the <see cref="ToMilliSeconds"/> and <see cref="ToByteOffset"/> methods.</remarks>
		/// <exception cref="ObjectDisposedException">The file has been closed.</exception>
		public int Length {
			get {
				if (m_Disposed)
					throw new ObjectDisposedException(this.GetType().FullName);
				return (int)(m_AudioLength / m_Format.nBlockAlign);
			}
		}
		/// <summary>
		/// Gets the volume of the media playback.
		/// </summary>
		/// <value>An integer representing the volume of the media playback.</value>
		/// <remarks>
		/// The low-order word of this integer contains the left-channel volume setting, and the high-order word contains the right-channel setting. A value of 0xFFFF represents full volume, and a value of 0x0000 is silence.
		/// <br>If a device does not support both left and right volume control, the low-order word of the specified location contains the mono volume level.</br>
		/// </remarks>
		/// <exception cref="ObjectDisposedException">The file has been closed.</exception>
		public int Volume {
			get {
				if (m_Disposed)
					throw new ObjectDisposedException(this.GetType().FullName);
				int volume;
				waveOutGetVolume(Output, out volume);
				return volume;
			}
		}
		/// <summary>
		/// Sets the volume of the media playback.
		/// </summary>
		/// <param name="volume">An integer representing the volume of the media playback. It must be an integer between 0x0 and 0xFFFF</param>
		/// <param name="channel">The channel to apply the change to.</param>
		/// <exception cref="ObjectDisposedException">The file has been closed.</exception>
		/// <exception cref="ArgumentOutOfRangeException"><paramref name="volume"/> is out of range.</exception>
		public void SetVolume(int volume, Channel channel) {
			if (m_Disposed)
				throw new ObjectDisposedException(this.GetType().FullName);
			if (volume < 0 || volume > 0xFFFF)
				throw new ArgumentOutOfRangeException();
			uint vol = (uint)volume;
			uint newvol = (uint)Volume;
			if (channel == Channel.Left || channel == Channel.Both) {
				newvol = vol | (newvol & 0xFFFF0000);
			}
			if (channel == Channel.Right || channel == Channel.Both) {
				newvol = (vol << 16) | (newvol & 0x0000FFFF);
			}
			waveOutSetVolume(Output, newvol);
		}
		/// <summary>
		/// Gets or sets the playback position of the file.
		/// </summary>
		/// <value>An Integer corresponding to the playback position of the file.</value>
		/// <remarks>You can convert between the byte offset of this property and milliseconds by using the <see cref="ToMilliSeconds"/> and <see cref="ToByteOffset"/> methods.</remarks>
		/// <exception cref="ObjectDisposedException">The file has been closed.</exception>
		public int Position {
			get {
				if (m_Disposed)
					throw new ObjectDisposedException(this.GetType().FullName);
				MMTIME tm = new MMTIME();
				tm.wType = TIME_BYTES;
				if (waveOutGetPosition(Output, ref tm, Marshal.SizeOf(tm)) == MMSYSERR_NOERROR)
					return (int)((m_StartPos + tm.u) / m_Format.nBlockAlign);
				else
					return (int)((mmioSeek(Input, 0, SEEK_CUR) - m_DataOffset + m_BufferSize * this.m_Buffers) / m_Format.nBlockAlign);
			}
			set {
				if (m_Disposed)
					throw new ObjectDisposedException(this.GetType().FullName);
				int bytepos = value * m_Format.nBlockAlign;
				mmioSeek(Input, bytepos + m_DataOffset, SEEK_SET);
			}
		}
		/// <summary>
		/// Gets the status of the media file.
		/// </summary>
		/// <value>One of the <see cref="MediaStatus"/> values.</value>
		public MediaStatus Status {
			get {
				return m_Status;
			}
		}
		/// <summary>
		/// Returns the multimedia input handle.
		/// </summary>
		/// <value>An <see cref="IntPtr"/> instance.</value>
		private IntPtr Input {
			get {
				return m_Input;
			}
			set {
				m_Input = value;
			}
		}
		/// <summary>
		/// Returns the WAVE output handle.
		/// </summary>
		/// <value>An <see cref="IntPtr"/> instance.</value>
		private IntPtr Output {
			get {
				return m_Output;
			}
			set {
				m_Output = value;
			}
		}
		/// <summary>
		/// Closes the media file and releases all unmanaged resources.
		/// </summary>
		public void Close() {
			if (!m_Disposed) {
				if (Status == MediaStatus.Playing)
					Stop();
				m_Disposed = true;
				mmioClose(Input, 0);
				for (int i = 0; i < m_Buffer.Length; i++) {
					Marshal.FreeHGlobal(m_Buffer[i]);
				}
				m_Status = MediaStatus.Closed;
				try {
					GC.SuppressFinalize(this);
				} catch {}
			}
		}
		/// <summary>
		/// Disposes of the media file.
		/// </summary>
		~WaveFile() {
			Close();
		}
		private string m_Filename;
		private bool m_Disposed;
		private IntPtr m_Input;
		private IntPtr m_Output;
		private int m_DataOffset;
		private int m_AudioLength;
		private int m_BufferSize;
		private int m_Buffers;
		private int m_StartPos;
		private int m_Stopping;
		private IntPtr[] m_Buffer;
		private WAVEHDR[] m_Header;
		private MediaStatus m_Status;
		private WAVEFORMAT m_Format;
		private MediaResult m_AsyncResult;
		// APIs
		private const int MMIO_READ = 0x0;
		private const int MMIO_FINDCHUNK = 0x10;
		private const int MMIO_FINDRIFF = 0x20;
		private const int SEEK_CUR = 1;
		private const int MMSYSERR_NOERROR = 0;
		private const float BUFFER_SECONDS = 0.1f;
		private const int CALLBACK_FUNCTION = 0x30000;
		private const int MM_WOM_DONE = 0x3BD;
		private const int SEEK_SET = 0;
		private const int TIME_BYTES = 0x4;

		[DllImport("winmm.dll", EntryPoint="mmioOpenA", CharSet=CharSet.Ansi)]
		private static extern IntPtr mmioOpen([MarshalAs(UnmanagedType.LPStr)] string szFileName, IntPtr lpmmioinfo, int dwOpenFlags);
		[DllImport("winmm.dll")]
		private static extern int mmioDescend(IntPtr hmmio, ref MMCKINFO lpck, IntPtr x, int uFlags);
		[DllImport("winmm.dll", EntryPoint="mmioStringToFOURCCA", CharSet=CharSet.Ansi)]
		private static extern int mmioStringToFOURCC([MarshalAs(UnmanagedType.LPStr)] string sz, int uFlags);
		[DllImport("winmm.dll")]
		private static extern int mmioClose(IntPtr hmmio, int uFlags);
		[DllImport("winmm.dll")]
		private static extern int mmioRead(IntPtr hmmio, ref WAVEFORMAT pch, int cch);
		[DllImport("winmm.dll")]
		private static extern int mmioRead(IntPtr hmmio, IntPtr pch, int cch);
		[DllImport("winmm.dll")]
		private static extern int mmioSeek(IntPtr hmmio, int lOffset, int iOrigin);
		[DllImport("winmm.dll")]
		private static extern int mmioAscend(IntPtr hmmio, ref MMCKINFO lpck, int uFlags);
		[DllImport("winmm.dll")]
		private static extern int waveOutReset(IntPtr hWaveIn);
		[DllImport("winmm.dll")]
		private static extern int waveOutUnprepareHeader(IntPtr hWaveIn, ref WAVEHDR lpWaveInHdr, int uSize);
		[DllImport("winmm.dll")]
		private static extern int waveOutClose(IntPtr hWaveIn);
		[DllImport("winmm.dll")]
		private static extern int waveOutWrite(IntPtr hWaveOut, ref WAVEHDR lpWaveOutHdr, int uSize);
		[DllImport("winmm.dll")]
		private static extern int waveOutOpen(ref IntPtr hWaveOut, int uDeviceID, ref WAVEFORMAT format, WaveDelegate dwCallback, int fPlaying, int dwFlags);
		[DllImport("winmm.dll")]
		private static extern int waveOutPrepareHeader(IntPtr hWaveIn, ref WAVEHDR lpWaveInHdr, int uSize);
		[DllImport("winmm.dll")]
		private static extern int waveOutGetPosition(IntPtr hWaveOut, ref MMTIME lpInfo, int uSize);
		[DllImport("winmm.dll")]
		private static extern int waveOutPause(IntPtr hWaveOut);
		[DllImport("winmm.dll")]
		private static extern int waveOutRestart(IntPtr hWaveOut);
		[DllImport("winmm.dll")]
		private static extern int waveOutSetVolume(IntPtr uDeviceID, uint dwVolume);
		[DllImport("winmm.dll")]
		private static extern int waveOutGetVolume(IntPtr uDeviceID, out int lpdwVolume);

		private delegate void WaveDelegate(IntPtr hwo, int uMsg, int dwInstance, ref WAVEHDR wavhdr, int dwParam2);
	}
	/// <summary>
	/// Defines the status values of the media file.
	/// </summary>
	public enum MediaStatus {
		/// <summary>The file is playing.</summary>
		Playing,
		/// <summary>The file is paused.</summary>
		Paused,
		/// <summary>The file is stopped.</summary>
		Stopped,
		/// <summary>The file is closed.</summary>
		Closed
	}
	/// <summary>
	/// Defines the different channels.
	/// </summary>
	public enum Channel {
		/// <summary>The left channel.</summary>
		Left,
		/// <summary>The right channel.</summary>
		Right,
		/// <summary>Both channels.</summary>
		Both
	}
	[StructLayout(LayoutKind.Sequential)]
	internal struct MMCKINFO {
		public int ckid;
		public int ckSize;
		public int fccType;
		public int dwDataOffset;
		public int dwFlags;
	}
	[StructLayout(LayoutKind.Sequential)]
	internal struct WAVEFORMAT {
		public short wFormatTag;
		public short nChannels;
		public int nSamplesPerSec;
		public int nAvgBytesPerSec;
		public short nBlockAlign;
		public short wBitsPerSample;
		public short cbSize;
	}
	[StructLayout(LayoutKind.Sequential)]
	internal struct WAVEHDR {
		public IntPtr lpData;
		public int dwBufferLength;
		public int dwBytesRecorded;
		public IntPtr dwUser;
		public int dwFlags;
		public int dwLoops;
		public IntPtr lpNext;
		public IntPtr Reserved;
	}
	[StructLayout(LayoutKind.Sequential)]
	internal struct MMTIME {
		public int wType;
		public int u;
		public int x;
	}
}