'
'    Copyright  2003, The KPD-Team
'    All rights reserved.
'    http:'www.mentalis.org/
'
'  Redistribution and use in source and binary forms, with or without
'  modification, are permitted provided that the following conditions
'  are met:
'
'    - Redistributions of source code must retain the above copyright
'       notice, this list of conditions and the following disclaimer. 
'
'    - Neither the name of the KPD-Team, nor the names of its contributors
'       may be used to endorse or promote products derived from this
'       software without specific prior written permission. 
'
'  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
'  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
'  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
'  FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL
'  THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
'  INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
'  (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
'  SERVICES LOSS OF USE, DATA, OR PROFITS OR BUSINESS INTERRUPTION)
'  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
'  STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
'  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
'  OF THE POSSIBILITY OF SUCH DAMAGE.
'
Option Explicit On 
Option Strict On

Imports System
Imports System.IO
Imports System.Text
Imports System.Runtime.InteropServices

Namespace Org.Mentalis.Multimedia
    ' <summary>
    ' Represents a WAVE sound file.
    ' </summary>
    Public NotInheritable Class WaveFile
        ' <summary>
        ' Initializes a new instance of the <see cref="WaveFile"/> class.
        ' </summary>
        ' <param name="file">The file to open.</param>
        ' <exception cref="MediaException">An error occurs while reading the media file.</exception>
        Public Sub New(ByVal file As String)
            Me.New(file, 5)
        End Sub
        ' <summary>
        ' Initializes a new instance of the <see cref="WaveFile"/> class.
        ' </summary>
        ' <param name="file">The file to open.</param>
        ' <param name="bufferCount">The number of memory buffers to use when reading the file.</param>
        ' <remarks>Setting the number of memory buffer to a low number may cause skips during playback. Setting the number of buffer to a high number may cause excessive memory usage.</remarks>
        ' <exception cref="MediaException">An error occurs while reading the media file.</exception>
        Public Sub New(ByVal file As String, ByVal bufferCount As Integer)
            If (file Is Nothing) Then Throw New ArgumentNullException()
            If (bufferCount <= 0) Then Throw New ArgumentException()
            m_Filename = file
            m_Disposed = False
            m_Buffers = bufferCount
            Open(file)
            m_Status = MediaStatus.Stopped
        End Sub
        ' <summary>
        ' Converts byte offsets returned by <see cref="Position"/> or <see cref="Length"/> to milliseconds.
        ' </summary>
        ' <param name="byteOffset">The byte offset the convert.</param>
        ' <returns>The corresponding milliseconds.</returns>
        Public Function ToMilliSeconds(ByVal byteOffset As Integer) As Integer
            Return CType(((byteOffset * m_Format.nBlockAlign) / m_Format.nAvgBytesPerSec) * 1000, Integer)
        End Function
        ' <summary>
        ' Converts byte offsets returned by <see cref="Position"/> or <see cref="Length"/> to a byte offset.
        ' </summary>
        ' <param name="milliseconds">The number of milliseconds to convert.</param>
        ' <returns>The corresponding byte offset.</returns>
        Public Function ToByteOffset(ByVal milliseconds As Integer) As Integer
            Return CType((CType(milliseconds / 1000, Integer) * m_Format.nAvgBytesPerSec) / m_Format.nBlockAlign, Integer)
        End Function
        ' <summary>
        ' Opens a WAVE file and initializes the internal structures.
        ' </summary>
        ' <param name="file">The file to open.</param>
        ' <exception cref="MediaException">An error occurs while reading the media file.</exception>
        Private Sub Open(ByVal filename As String)
            ' Make sure the file exists
            If (Not File.Exists(filename)) Then Throw New FileNotFoundException("The media file could not be found.", filename)
            ' Open the input file
            Input = mmioOpen(filename, IntPtr.Zero, MMIO_READ)
            If (Input.Equals(IntPtr.Zero)) Then Throw New MediaException("An error occurs while opening the media file.")
            ' make sure this is a wave file
            Dim mmi As New MMCKINFO()
            mmi.fccType = mmioStringToFOURCC("WAVE", 0)
            If (mmioDescend(Input, mmi, IntPtr.Zero, MMIO_FINDRIFF) <> MMSYSERR_NOERROR) Then
                Close()
                Throw New MediaException("The media file is invalid.")
            End If
            ' get format info
            mmi = New MMCKINFO()
            mmi.ckid = mmioStringToFOURCC("fmt", 0)
            If (mmioDescend(Input, mmi, IntPtr.Zero, MMIO_FINDCHUNK) <> MMSYSERR_NOERROR) Then
                Close()
                Throw New MediaException("The format chunk could not be found.")
            End If
            m_Format = New WAVEFORMAT()
            If (mmioRead(Input, m_Format, Marshal.SizeOf(m_Format)) = -1) Then
                Close()
                Throw New MediaException("An error occurs while reading from the media file.")
            End If
            mmioAscend(Input, mmi, 0)
            ' find the data subchunk
            mmi.ckid = mmioStringToFOURCC("data", 0)
            If (mmioDescend(Input, mmi, IntPtr.Zero, MMIO_FINDCHUNK) <> MMSYSERR_NOERROR) Then
                Close()
                Throw New MediaException("The data chunk could not be found.")
            End If
            m_DataOffset = mmioSeek(Input, 0, SEEK_CUR)
            ' Get the length of the audio
            m_AudioLength = mmi.ckSize
            ' Allocate audio buffers
            m_BufferSize = CType(m_Format.nSamplesPerSec * m_Format.nBlockAlign * m_Format.nChannels * BUFFER_SECONDS, Integer)
            m_BufferSize = m_BufferSize - (m_BufferSize Mod m_Format.nBlockAlign)
            ReDim m_Buffer(m_Buffers - 1)
            ReDim m_Header(m_Buffers - 1)
            Dim i As Integer
            For i = 0 To m_Buffer.Length - 1
                m_Buffer(i) = Marshal.AllocHGlobal(m_BufferSize)
            Next
        End Sub
        ' <summary>
        ' Starts playing the media file synchronously.
        ' </summary>
        ' <remarks>This method will block until the file has finished playing.</remarks>
        ' <exception cref="ObjectDisposedException">The file has been closed.</exception>
        ' <exception cref="MediaException">An error occurs while playing the media file.</exception>
        Public Sub Play()
            Dim ar As IAsyncResult = BeginPlay(Nothing, Nothing)
            ar.AsyncWaitHandle.WaitOne()
        End Sub
        ' <summary>
        ' Starts playing the media file asynchronously.
        ' </summary>
        ' <param name="callback">An <see cref="AsyncCallback"/> instance. This delegate will be called if the playback is finished. This member can be a null reference (<b>Nothing</b> in Visual Basic).</param>
        ' <param name="state">An object containing state information for this request.</param>
        ' <returns>An IAsyncResult that references the asynchronous playback.</returns>
        ' <exception cref="ObjectDisposedException">The file has been closed.</exception>
        ' <exception cref="MediaException">An error occurs while playing the media file.</exception>
        Public Function BeginPlay(ByVal callback As AsyncCallback, ByVal state As Object) As IAsyncResult
            If (m_Disposed) Then Throw New ObjectDisposedException(Me.GetType().FullName)
            If (m_Status = MediaStatus.Playing) Then Return m_AsyncResult
            If (waveOutOpen(m_Output, 0, m_Format, New WaveDelegate(AddressOf WaveCallBack), 0, CALLBACK_FUNCTION) <> MMSYSERR_NOERROR) Then Throw New MediaException("An error occurs while opening the WAVE device.")
            Dim i As Integer
            For i = 0 To m_Header.Length - 1
                m_Header(i).lpData = m_Buffer(i)
                m_Header(i).dwBufferLength = m_BufferSize
                m_Header(i).dwFlags = 0
                m_Header(i).dwLoops = 0
                If (waveOutPrepareHeader(Output, m_Header(i), Marshal.SizeOf(m_Header(i))) <> MMSYSERR_NOERROR) Then Throw New MediaException("An error occurs while preparing the WAVE buffers.")
            Next
            m_Stopping = -1
            m_Status = MediaStatus.Playing
            m_StartPos = mmioSeek(Input, 0, SEEK_CUR) - m_DataOffset
            For i = 0 To m_Header.Length - 1
                WaveCallBack(Input, MM_WOM_DONE, 0, m_Header(i), 0)
            Next
            m_AsyncResult = New MediaResult(callback, state)
            Return m_AsyncResult
        End Function
        ' <summary>
        ' Called whenever a data buffer has been played.
        ' </summary>
        ' <param name="hwo">Handle to the waveform-audio device associated with the callback.</param>
        ' <param name="uMsg">Waveform-audio output message.</param>
        ' <param name="dwInstance">User-instance data specified with waveOutOpen.</param>
        ' <param name="wavhdr">Message parameter.</param>
        ' <param name="dwParam2">Message parameter.</param>
        Private Sub WaveCallBack(ByVal hwo As IntPtr, ByVal uMsg As Integer, ByVal dwInstance As Integer, ByRef wavhdr As WAVEHDR, ByVal dwParam2 As Integer)
            If (uMsg = MM_WOM_DONE AndAlso m_Stopping = -1) Then
                Dim rc As Integer
                Dim remaining As Integer = m_DataOffset + m_AudioLength - mmioSeek(Input, 0, SEEK_CUR)
                If (m_BufferSize < remaining) Then
                    rc = mmioRead(Input, wavhdr.lpData, m_BufferSize)
                Else
                    rc = mmioRead(Input, wavhdr.lpData, remaining)
                    m_Stopping = 0
                End If
                If (rc <> -1) Then
                    wavhdr.dwBufferLength = rc
                    rc = waveOutWrite(Output, wavhdr, Marshal.SizeOf(wavhdr))
                End If
            End If
            If (uMsg = MM_WOM_DONE AndAlso m_Stopping <> -1) Then
                m_Stopping += 1
                If (m_Stopping = m_Buffers + 1) Then
                    Me.Position = 0
                    m_Status = MediaStatus.Stopped
                    m_AsyncResult.Notify()
                End If
            End If
        End Sub
        ' <summary>
        ' Stops the playback of the file.
        ' </summary>
        ' <exception cref="ObjectDisposedException">The file has been closed.</exception>
        Public Sub StopPlay()
            If (m_Disposed) Then Throw New ObjectDisposedException(Me.GetType().FullName)
            If (m_Status = MediaStatus.Stopped) Then Return
            m_Stopping = 1
            waveOutReset(Output)
            Dim i As Integer
            For i = 0 To m_Header.Length - 1
                waveOutUnprepareHeader(Output, m_Header(i), Marshal.SizeOf(m_Header(i)))
            Next
            waveOutClose(Output)
            Position = 0
            m_Status = MediaStatus.Stopped
        End Sub
        ' <summary>
        ' Pauses or unpauses the playback of the file.
        ' </summary>
        ' <value><b>true</b> if the file is paused, <b>false</b> otherwise.</value>
        ' <exception cref="ObjectDisposedException">The file has been closed.</exception>
        Public Property Paused() As Boolean
            Get
                If (m_Disposed) Then Throw New ObjectDisposedException(Me.GetType().FullName)
                Return Me.Status = MediaStatus.Paused
            End Get
            Set(ByVal Value As Boolean)
                If (m_Disposed) Then Throw New ObjectDisposedException(Me.GetType().FullName)
                If (Me.Status = MediaStatus.Stopped) Then Return
                If (Value AndAlso Me.Status = MediaStatus.Playing) Then
                    waveOutPause(Output)
                    m_Status = MediaStatus.Paused
                ElseIf (Not Value AndAlso Me.Status = MediaStatus.Paused) Then
                    waveOutRestart(Output)
                    m_Status = MediaStatus.Playing
                End If
            End Set
        End Property
        ' <summary>
        ' Gets the full path of the media file.
        ' </summary>
        ' <value>A <see cref="String"/>.</value>
        Public ReadOnly Property Filename() As String
            Get
                Return m_Filename
            End Get
        End Property
        ' <summary>
        ' Gets the length of the media file.
        ' </summary>
        ' <value>An Integer representing the length of the file.</value>
        ' <remarks>You can convert between the byte offset of this property and milliseconds by using the <see cref="ToMilliSeconds"/> and <see cref="ToByteOffset"/> methods.</remarks>
        ' <exception cref="ObjectDisposedException">The file has been closed.</exception>
        Public ReadOnly Property Length() As Integer
            Get
                If (m_Disposed) Then Throw New ObjectDisposedException(Me.GetType().FullName)
                Return CType(m_AudioLength / m_Format.nBlockAlign, Integer)
            End Get
        End Property
        ' <summary>
        ' Gets the volume of the media playback.
        ' </summary>
        ' <value>An integer representing the volume of the media playback.</value>
        ' <remarks>
        ' The low-order word of this integer contains the left-channel volume setting, and the high-order word contains the right-channel setting. A value of 0xFFFF represents full volume, and a value of 0x0000 is silence.
        ' <br>If a device does not support both left and right volume control, the low-order word of the specified location contains the mono volume level.</br>
        ' </remarks>
        ' <exception cref="ObjectDisposedException">The file has been closed.</exception>
        Public ReadOnly Property Volume() As Integer
            Get
                If (m_Disposed) Then Throw New ObjectDisposedException(Me.GetType().FullName)
                Dim vol As Integer
                waveOutGetVolume(Output, vol)
                Return vol
            End Get
        End Property
        ' <summary>
        ' Sets the volume of the media playback.
        ' </summary>
        ' <param name="volume">An integer representing the volume of the media playback. It must be an integer between 0x0 and 0xFFFF</param>
        ' <param name="channel">The channel to apply the change to.</param>
        ' <exception cref="ObjectDisposedException">The file has been closed.</exception>
        ' <exception cref="ArgumentOutOfRangeException"><paramref name="volume"/> is out of range.</exception>
        Public Sub SetVolume(ByVal volume As Integer, ByVal Channel As Channel)
            If (m_Disposed) Then Throw New ObjectDisposedException(Me.GetType().FullName)
            If (volume < 0 OrElse volume > &HFFFF) Then Throw New ArgumentOutOfRangeException()
            Dim newvol As Long = volume
            If (Channel = Channel.Left OrElse Channel = Channel.Both) Then
                newvol = volume Or (newvol And &HFFFF0000)
            End If
            If (Channel = Channel.Right OrElse Channel = Channel.Both) Then
                newvol = CType(volume * (2 ^ 16), Long) Or (newvol And &HFFFF)
            End If
            waveOutSetVolume(Output, CType(newvol, Integer))
        End Sub
        ' <summary>
        ' Gets or sets the playback position of the file.
        ' </summary>
        ' <value>An Integer corresponding to the playback position of the file.</value>
        ' <remarks>You can convert between the byte offset of this property and milliseconds by using the <see cref="ToMilliSeconds"/> and <see cref="ToByteOffset"/> methods.</remarks>
        ' <exception cref="ObjectDisposedException">The file has been closed.</exception>
        Public Property Position() As Integer
            Get
                If (m_Disposed) Then Throw New ObjectDisposedException(Me.GetType().FullName)
                Dim tm As New MMTIME()
                tm.wType = TIME_BYTES
                If (waveOutGetPosition(Output, tm, Marshal.SizeOf(tm)) = MMSYSERR_NOERROR) Then
                    Return CType((m_StartPos + tm.u) / m_Format.nBlockAlign, Integer)
                Else
                    Return CType((mmioSeek(Input, 0, SEEK_CUR) - m_DataOffset + m_BufferSize * Me.m_Buffers) / m_Format.nBlockAlign, Integer)
                End If
            End Get
            Set(ByVal Value As Integer)
                If (m_Disposed) Then Throw New ObjectDisposedException(Me.GetType().FullName)
                Dim bytepos As Integer = Value * m_Format.nBlockAlign
                mmioSeek(Input, bytepos + m_DataOffset, SEEK_SET)
            End Set
        End Property
        ' <summary>
        ' Gets the status of the media file.
        ' </summary>
        ' <value>One of the <see cref="MediaStatus"/> values.</value>
        Public ReadOnly Property Status() As MediaStatus
            Get
                Return m_Status
            End Get
        End Property
        ' <summary>
        ' Returns the multimedia input handle.
        ' </summary>
        ' <value>An <see cref="IntPtr"/> instance.</value>
        Private Property Input() As IntPtr
            Get
                Return m_Input
            End Get
            Set(ByVal Value As IntPtr)
                m_Input = Value
            End Set
        End Property
        ' <summary>
        ' Returns the WAVE output handle.
        ' </summary>
        ' <value>An <see cref="IntPtr"/> instance.</value>
        Private Property Output() As IntPtr
            Get
                Return m_Output
            End Get
            Set(ByVal Value As IntPtr)
                m_Output = Value
            End Set
        End Property
        ' <summary>
        ' Closes the media file and releases all unmanaged resources.
        ' </summary>
        Public Sub Close()
            If (Not m_Disposed) Then
                If Status = MediaStatus.Playing Then StopPlay()
                m_Disposed = True
                mmioClose(Input, 0)
                Dim i As Integer
                For i = 0 To m_Buffer.Length - 1
                    Marshal.FreeHGlobal(m_Buffer(i))
                Next
                m_Status = MediaStatus.Closed
                Try
                    GC.SuppressFinalize(Me)
                Catch
                End Try
            End If
        End Sub
        ' <summary>
        ' Disposes of the media file.
        ' </summary>
        Protected Overrides Sub Finalize()
            Close()
        End Sub
        Private m_Filename As String
        Private m_Disposed As Boolean
        Private m_Input As IntPtr
        Private m_Output As IntPtr
        Private m_DataOffset As Integer
        Private m_AudioLength As Integer
        Private m_BufferSize As Integer
        Private m_Buffers As Integer
        Private m_StartPos As Integer
        Private m_Stopping As Integer
        Private m_Buffer() As IntPtr
        Private m_Header() As WAVEHDR
        Private m_Status As MediaStatus
        Private m_Format As WAVEFORMAT
        Private m_AsyncResult As MediaResult
        ' APIs
        Private Const MMIO_READ As Integer = &H0
        Private Const MMIO_FINDCHUNK As Integer = &H10
        Private Const MMIO_FINDRIFF As Integer = &H20
        Private Const SEEK_CUR As Integer = 1
        Private Const MMSYSERR_NOERROR As Integer = 0
        Private Const BUFFER_SECONDS As Single = 0.1F
        Private Const CALLBACK_FUNCTION As Integer = &H30000
        Private Const MM_WOM_DONE As Integer = &H3BD
        Private Const SEEK_SET As Integer = 0
        Private Const TIME_BYTES As Integer = &H4

        Private Declare Ansi Function mmioOpen Lib "winmm.dll" Alias "mmioOpenA" (<MarshalAs(UnmanagedType.LPStr)> ByVal szFileName As String, ByVal lpmmioinfo As IntPtr, ByVal dwOpenFlags As Integer) As IntPtr
        Private Declare Function mmioDescend Lib "winmm.dll" (ByVal hmmio As IntPtr, ByRef lpck As MMCKINFO, ByVal x As IntPtr, ByVal uFlags As Integer) As Integer
        Private Declare Ansi Function mmioStringToFOURCC Lib "winmm.dll" Alias "mmioStringToFOURCCA" (<MarshalAs(UnmanagedType.LPStr)> ByVal sz As String, ByVal uFlags As Integer) As Integer
        Private Declare Function mmioClose Lib "winmm.dll" (ByVal hmmio As IntPtr, ByVal uFlags As Integer) As Integer
        Private Declare Function mmioRead Lib "winmm.dll" (ByVal hmmio As IntPtr, ByRef pch As WAVEFORMAT, ByVal cch As Integer) As Integer
        Private Declare Function mmioRead Lib "winmm.dll" (ByVal hmmio As IntPtr, ByVal pch As IntPtr, ByVal cch As Integer) As Integer
        Private Declare Function mmioSeek Lib "winmm.dll" (ByVal hmmio As IntPtr, ByVal lOffset As Integer, ByVal iOrigin As Integer) As Integer
        Private Declare Function mmioAscend Lib "winmm.dll" (ByVal hmmio As IntPtr, ByRef lpck As MMCKINFO, ByVal uFlags As Integer) As Integer
        Private Declare Function waveOutReset Lib "winmm.dll" (ByVal hWaveIn As IntPtr) As Integer
        Private Declare Function waveOutUnprepareHeader Lib "winmm.dll" (ByVal hWaveIn As IntPtr, ByRef lpWaveInHdr As WAVEHDR, ByVal uSize As Integer) As Integer
        Private Declare Function waveOutClose Lib "winmm.dll" (ByVal hWaveIn As IntPtr) As Integer
        Private Declare Function waveOutWrite Lib "winmm.dll" (ByVal hWaveOut As IntPtr, ByRef lpWaveOutHdr As WAVEHDR, ByVal uSize As Integer) As Integer
        Private Declare Function waveOutOpen Lib "winmm.dll" (ByRef hWaveOut As IntPtr, ByVal uDeviceID As Integer, ByRef format As WAVEFORMAT, ByVal dwCallback As WaveDelegate, ByVal fPlaying As Integer, ByVal dwFlags As Integer) As Integer
        Private Declare Function waveOutPrepareHeader Lib "winmm.dll" (ByVal hWaveIn As IntPtr, ByRef lpWaveInHdr As WAVEHDR, ByVal uSize As Integer) As Integer
        Private Declare Function waveOutGetPosition Lib "winmm.dll" (ByVal hWaveOut As IntPtr, ByRef lpInfo As MMTIME, ByVal uSize As Integer) As Integer
        Private Declare Function waveOutPause Lib "winmm.dll" (ByVal hWaveOut As IntPtr) As Integer
        Private Declare Function waveOutRestart Lib "winmm.dll" (ByVal hWaveOut As IntPtr) As Integer
        Private Declare Function waveOutSetVolume Lib "winmm.dll" (ByVal uDeviceID As IntPtr, ByVal dwVolume As Integer) As Integer
        Private Declare Function waveOutGetVolume Lib "winmm.dll" (ByVal uDeviceID As IntPtr, ByRef lpdwVolume As Integer) As Integer

        Private Delegate Sub WaveDelegate(ByVal hwo As IntPtr, ByVal uMsg As Integer, ByVal dwInstance As Integer, ByRef wavhdr As WAVEHDR, ByVal dwParam2 As Integer)
    End Class
    ' <summary>
    ' Defines the status values of the media file.
    ' </summary>
    Public Enum MediaStatus
        ' <summary>The file is playing.</summary>
        Playing
        ' <summary>The file is paused.</summary>
        Paused
        ' <summary>The file is stopped.</summary>
        Stopped
        ' <summary>The file is closed.</summary>
        Closed
    End Enum
    ' <summary>
    ' Defines the different channels.
    ' </summary>
    Public Enum Channel
        ' <summary>The left channel.</summary>
        Left
        ' <summary>The right channel.</summary>
        Right
        ' <summary>Both channels.</summary>
        Both
    End Enum
    <StructLayout(LayoutKind.Sequential)> _
    Friend Structure MMCKINFO
        Public ckid As Integer
        Public ckSize As Integer
        Public fccType As Integer
        Public dwDataOffset As Integer
        Public dwFlags As Integer
    End Structure
    <StructLayout(LayoutKind.Sequential)> _
    Friend Structure WAVEFORMAT
        Public wFormatTag As Short
        Public nChannels As Short
        Public nSamplesPerSec As Integer
        Public nAvgBytesPerSec As Integer
        Public nBlockAlign As Short
        Public wBitsPerSample As Short
        Public cbSize As Short
    End Structure
    <StructLayout(LayoutKind.Sequential)> _
           Friend Structure WAVEHDR
        Public lpData As IntPtr
        Public dwBufferLength As Integer
        Public dwBytesRecorded As Integer
        Public dwUser As IntPtr
        Public dwFlags As Integer
        Public dwLoops As Integer
        Public lpNext As IntPtr
        Public Reserved As IntPtr
    End Structure
    <StructLayout(LayoutKind.Sequential)> _
    Friend Structure MMTIME
        Public wType As Integer
        Public u As Integer
        Public x As Integer
    End Structure
End Namespace